#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>

#include <sys/types.h>
#include <termios.h>

#ifndef TIOCGWINSZ
# include <sys/ioctl.h>   /* 44BSD requires this too */
#endif

#include "pty.h"

int ptym_open(char *pts_name)
{
#ifdef __BEOS__
   strcpy(pts_name, "/dev/pt/");
#else
   strcpy(pts_name, "/dev/pty");
#endif

   int base = strlen(pts_name);

   /* array index: 0123456789 (for references in following code) */
   for (const char * ptr1 = "pqrstuvwxyzPQRST"; *ptr1 != 0; ptr1++) 
   {
      pts_name[base] = *ptr1;
      for (const char * ptr2 = "0123456789abcdef"; *ptr2 != 0; ptr2++) 
      {
         pts_name[base+1] = *ptr2;
         pts_name[base+2] = '\0';

         /* try to open master */
         int fdm = open(pts_name, O_RDWR);
         if (fdm < 0) 
         {
            if (errno == ENOENT) return -1; /* out of pty devices */
                            else continue;  /* try next pty device */
         }

         pts_name[5] = 't';  /* change "pty" to "tty" */
         return fdm ;        /* got it, return fd of master */
      }
   }
   return -1;      /* out of pty devices */
}

int ptys_open(int fdm, char *pts_name)
{
   struct group *grptr = getgrnam("tty");

   /* following two functions don't work unless we're root */
   chown(pts_name, getuid(), grptr?grptr->gr_gid:(gid_t)-1);
   chmod(pts_name, S_IRUSR | S_IWUSR | S_IWGRP);

   int fds = open(pts_name, O_RDWR);
   if (fds < 0) 
   {
      close(fdm);
      return(-1);
   }
   return(fds);
}

pid_t pty_fork(int *ptrfdm, char *slave_name, const struct termios *slave_termios, const struct winsize *slave_winsize)
{
   char  pts_name[20];

   int fdm = ptym_open(pts_name);
   if (fdm < 0) fprintf(stderr, "can't open master pty %s: (%s)\n", pts_name, strerror(errno));
   if (slave_name) strcpy(slave_name, pts_name);   /* return name of slave */

   pid_t pid = fork();
        if (pid < 0) return(-1);
   else if (pid == 0) 
   {
      /* child */
      if (setsid() < 0) fprintf(stderr, "setsid error: %s\n", strerror(errno));

      /* SVR4 acquires controlling terminal on open() */
      int fds = ptys_open(fdm, pts_name);
      if (fds < 0) fprintf(stderr, "can't open slave pty: %s\n", strerror(errno));
      close(fdm);      /* all done with master in child */

#if defined(TIOCSCTTY) && !defined(CIBAUD)
      /* 44BSD way to acquire controlling terminal */
      /* !CIBAUD to avoid doing this under SunOS */
      if (ioctl(fds, TIOCSCTTY, (char *) 0) < 0) fprintf(stderr, "TIOCSCTTY error: %s\n", strerror(errno));
#endif

      /* set slave's termios and window size */
      if ((slave_termios)&&(tcsetattr(fds, TCSANOW, slave_termios) < 0)) fprintf(stderr, "tcsetattr error on slave pty: %s\n", strerror(errno));
      if ((slave_winsize)&&(ioctl(fds, TIOCSWINSZ,  slave_winsize) < 0)) fprintf(stderr, "TIOCSWINSZ error on slave pty: %s\n", strerror(errno));

      /* slave becomes stdin/stdout/stderr of child */
      if (dup2(fds, STDIN_FILENO)  != STDIN_FILENO)  fprintf(stderr, "dup2 error to stdin: %s\n", strerror(errno));
      if (dup2(fds, STDOUT_FILENO) != STDOUT_FILENO) fprintf(stderr, "dup2 error to stdout: %s\n", strerror(errno));
      if (fds > STDERR_FILENO) close(fds);
      return(0);      /* child returns 0 just like fork() */
   } 
   else 
   {
      /* parent */
      *ptrfdm = fdm; /* return fd of master */
      return(pid);   /* parent returns pid of child */
   }
}
