#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <termios.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <time.h>
#include <unistd.h>

#include "muscle/dataio/FileDescriptorDataIO.h"
#include "muscle/dataio/TCPSocketDataIO.h"
#include "muscle/iogateway/MessageIOGateway.h"
#include "muscle/iogateway/PlainTextMessageIOGateway.h"
#include "muscle/reflector/StorageReflectConstants.h"
#include "muscle/regex/PathMatcher.h"
#include "muscle/system/SetupSystem.h"
#include "muscle/util/MiscUtilityFunctions.h"
#include "muscle/util/NetworkUtilityFunctions.h"
#include "muscle/util/StringTokenizer.h"

#include "pty.h"

using namespace muscle;

static const uint32 NET_CLIENT_NEW_CHAT_TEXT = 2;

// Launches a subprocess with the given command line, and returns a file descriptor
// that can be used to send to and receive some the sub-process.
static int OpenSubprocess(const char * cmd, bool blocking)
{
   struct termios orig_termios;
   struct winsize size;
   if (tcgetattr(STDIN_FILENO, &orig_termios) < 0) fprintf(stderr, "tcgetattr error on stdin: %s\n", strerror(errno));
   if (ioctl(STDIN_FILENO, TIOCGWINSZ, (char *) &size) < 0) fprintf(stderr, "TIOCGWINSZ error: %s\n", strerror(errno));

   int ptySock;
   pid_t pid = pty_fork(&ptySock, NULL, &orig_termios, &size);
   if (pid < 0) return -1;  // fork failure!
   else if (pid == 0)
   {
      // we are the child process -- turn off echo and execute the command
      struct termios stermios; 
      if (tcgetattr(STDIN_FILENO, &stermios) < 0) fprintf(stderr, "tcgetattr error: %s\n", strerror(errno)); 
      stermios.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL); 
      stermios.c_oflag &= ~(ONLCR); /* also turn off NL to CR/NL mapping on output */ 
      if (tcsetattr(STDIN_FILENO, TCSANOW, &stermios) < 0) fprintf(stderr, "tcsetattr error: %s\n", strerror(errno)); 
      _exit(system(cmd));
   }
   else
   {
      // we are the parent process
      SetSocketBlockingEnabled(ptySock, blocking);
      return ptySock;
   }
}

class ShellBot : public AbstractGatewayMessageReceiver
{
public:
   ShellBot(const Message & args) : _args(args)
   {
      _botName = "shellbot"; (void) _args.FindString("name", _botName);

      String command;
      String hostName;
      uint16 port = 2960;
      if ((_args.FindString("command", command) == B_NO_ERROR)&&(ParseConnectArg(args, "host", hostName, port) == B_NO_ERROR))
      {
         LogTime(MUSCLE_LOG_INFO, "Arguments: name=[%s] command=[%s]\n", _botName(), command());
         bool keepGoing = true;
         while(keepGoing)
         {
            int subSocket = OpenSubprocess(command.Trim()(), false);
            if (subSocket >= 0)
            {
               _subGateway.Reset(); 
               _subGateway.SetDataIO(DataIORef(new FileDescriptorDataIO(subSocket, false), NULL));
               _subGateway.SetOutgoingEndOfLineString("\n");

               int netSocket = Connect(hostName(), port, "shellbot", false);
               if (netSocket >= 0)
               {
                  _netGateway.Reset(); 
                  _netGateway.SetDataIO(DataIORef(new TCPSocketDataIO(netSocket, false), NULL));

                  // Upload our user info
                  {
                     Message infoMsg;
                     infoMsg.AddString("name", _botName);
                     infoMsg.AddInt32("port",  0);
                     infoMsg.AddBool("bot",    true);

                     MessageRef setNameMessage = GetMessageFromPool(PR_COMMAND_SETDATA);
                     if (setNameMessage())
                     {
                        setNameMessage()->AddMessage("beshare/name", infoMsg);
                        _netGateway.AddOutgoingMessage(setNameMessage); 
                     }
                  }

                  // Then handle incoming commands, etc
                  while(true)
                  {
                     fd_set readSet;  FD_ZERO(&readSet);
                     fd_set writeSet; FD_ZERO(&writeSet);

                     FD_SET(netSocket, &readSet);
                     FD_SET(subSocket, &readSet);
                     if (_netGateway.HasBytesToOutput()) FD_SET(netSocket, &writeSet);
                     if (_subGateway.HasBytesToOutput()) FD_SET(subSocket, &writeSet);
                  
                     if (select(muscleMax(netSocket, subSocket)+1, &readSet, &writeSet, NULL, NULL) < 0)
                     {
                        LogTime(MUSCLE_LOG_CRITICALERROR, "shellbot: select() failed, exiting!\n");
                        keepGoing = false;
                        break;
                     }

                     // handle subprocess I/O
                     {
                        _handlingNetworkIO = false;
                        bool writing = FD_ISSET(subSocket, &writeSet);
                        bool reading = FD_ISSET(subSocket, &readSet);
                        bool writeError = ((writing)&&(_subGateway.DoOutput()     < 0));
                        bool readError  = ((reading)&&(_subGateway.DoInput(*this) < 0));
                        if ((readError)||(writeError))
                        {
                           LogTime(MUSCLE_LOG_ERROR, "Connection to subprocess closed, exiting!\n");
                           keepGoing = false;
                           break;
                        }
                     }

                     // handle network I/O
                     {
                        _handlingNetworkIO = true;
                        bool writing = FD_ISSET(netSocket, &writeSet);
                        bool reading = FD_ISSET(netSocket, &readSet);
                        bool writeError = ((writing)&&(_netGateway.DoOutput()     < 0));
                        bool readError  = ((reading)&&(_netGateway.DoInput(*this) < 0));
                        if ((readError)||(writeError))
                        {
                           LogTime(MUSCLE_LOG_ERROR, "Connection to server closed, reconnecting...\n");
                           break;
                        }
                     }
                  }
               }
               if (keepGoing)
               { 
                  LogTime(MUSCLE_LOG_INFO, "Will auto-restart in 2 seconds.\n");
                  Snooze64(2000000);
               }
            }
            else
            {
               LogTime(MUSCLE_LOG_CRITICALERROR, "Error creating child process [%s]\n", command.Trim()());
               break;
            }
         }
      }
      else LogTime(MUSCLE_LOG_CRITICALERROR, "Usage:  shellbot host=beshare.tycomsystems.com command=\"shell command args\" [name=botname]\n");
   }

   virtual void MessageReceivedFromGateway(MessageRef msg, void *)
   {
      if (_handlingNetworkIO)
      {
         String text;
         String from;
         if ((msg()->what == NET_CLIENT_NEW_CHAT_TEXT)&&
             (msg()->HasName("bot")              == false)&&
             (msg()->FindString("text", text)    == B_NO_ERROR)&&
             (msg()->FindString("session", from) == B_NO_ERROR))
         {
            text = text.Trim();
   
                 if (msg()->HasName("private")) SendChatMessage("Sorry, but I can't handle private messages!", from, true);
            else if (text.StartsWithIgnoreCase(_botName))
            {
               text = text.Substring(_botName.Length()).Trim();
               if (text.StartsWithIgnoreCase("owner")) 
               {
                  const char * owner = "<unknown>"; (void) _args.FindString("owner", &owner);
                  SendChatMessage(String("My owner is %1").Arg(owner), "*", false);
               }

               MessageRef toSubProcess = GetMessageFromPool(PR_COMMAND_TEXT_STRINGS);
               if ((toSubProcess())&&(toSubProcess()->AddString(PR_NAME_TEXT_LINE, text) == B_NO_ERROR)) _subGateway.AddOutgoingMessage(toSubProcess);
            }
         }
      }
      else
      {
         bool firstTime = true;
         String text, line;
         for (uint32 i=0; msg()->FindString(PR_NAME_TEXT_LINE, i, line) == B_NO_ERROR; i++) 
         {
            // If it's going to be multiline, add an initial carriage return so
            // that the first line isn't offset from the other lines.
            if ((firstTime)&&(text.IndexOf('\n') >= 0))
            {
               text = text.Prepend("\n");
               firstTime = false;
            }
            text += line;
         }
         if (text.Length() > 0) SendChatMessage(text, "*", false);
      }
   }

private:
   void SendChatMessage(const String & text, const String & target, bool priv)
   {
      MessageRef msg = GetMessageFromPool(NET_CLIENT_NEW_CHAT_TEXT);
      if ((msg())&&
          (msg()->AddString(PR_NAME_KEYS, String("/*/%1/beshare").Arg(target)) == B_NO_ERROR)&&
          (msg()->AddString("session", "") == B_NO_ERROR)&&
          (msg()->AddString("text", text)  == B_NO_ERROR)&&
          (msg()->AddBool("bot", true)     == B_NO_ERROR)&&
          ((priv == false)||(msg()->AddBool("private", true) == B_NO_ERROR))) _netGateway.AddOutgoingMessage(msg);
   }

   const Message & _args;
   MessageIOGateway _netGateway;
   PlainTextMessageIOGateway _subGateway;
   String _botName;
   bool _handlingNetworkIO;
};

int main(int argc, char ** argv)
{
   muscle::CompleteSetupSystem css;
   Message args; (void) ParseArgs(argc, argv, args);
   HandleStandardDaemonArgs(args);
   ShellBot shellBot(args);  // everything happens in the ctor!
   return 0;
}
